#
# Copyright Centeris Corporation.  All rights reserved.
#

use strict;
use warnings;

use Getopt::Long;
use File::Basename;
use File::Spec;
use File::Path;
use FindBin;
use Config;
use IO::File;
use Cwd;

STDOUT->autoflush(1);
STDERR->autoflush(1);

my $TEST_MODE;
my $ID_PATTERN = "[A-Za-z_][A-Za-z_.0-9]*";

my @REQUIRED_OPTIONS_KEYS = qw(name number repository revision);
my @OPTIONAL_OPTIONS_KEYS = qw(user scp_command ssh_command noprepare);
my @VALID_OPTIONS_KEYS = (@REQUIRED_OPTIONS_KEYS, @OPTIONAL_OPTIONS_KEYS);

my @REQUIRED_PHASE_KEYS = qw(machine root dist);
my @VALID_PHASE_KEYS = (qw(inherit depends path command checkdist sparse scpdist cond noprepare nobuildenv image meta run cache), @REQUIRED_PHASE_KEYS);

my $SVN = 'svn';

my $STATUS_SKIPPED = 'SKIPPED';
my $STATUS_DONE = 'DONE';
my $STATUS_ERROR = 'ERROR';

my $SUB_STATUS_DEP = 'DEP';

sub main();
exit(main());


sub usage()
{
    my $scriptName = fileparse($0);

    return <<DATA;
usage: $scriptName [options]

    This does the bulk of the work for the official build.  It is invoked by
    official-build shell script wrapper after the wrapper has set up the build
    environment for the official build.

    The assumptions are:

    1) Perl exists on the system and is in the PATH.

    2) Subversion exists on the system and is in the PATH.

  Options:

    --help            Show this help information.

    --test            Do not run any commands, just output what would be run.

    --revert          Do svn revert when svn updating.  (RECOMMENDED)

  One of these is required:

    --config FILE     Configuration file to use.

    --phase PHASE     Indicate what phase we are at (internal option).

  Required:

    --name NAME       The name of the code base (diagnostics).

    --number NUM      The build number (for buildenv).

    --repository URL  The svn repository URL.

    --revision REV    The svn revision.

  One of these is required with --config:

    --log PATH        Log file base name.  The phase name is appended for each
                      child phase spawned.

    --nospawn         Do not spawn children for each phase.  The default
                      is to spawn children for each phase, with independent
                      phases able to run concurrently.

  Optional with --config:

    --define VAR=VAL  Define a variable to use in the config file.

    --user USER       The username to build as (also passed to buildenv).

    --ssh_command     Command to use for building remotely (defaults to
                      what is in .conf file).

    --scp_command     Command to use for copying files remotely (defaults to
                      what is in .conf file).

  Required when using --phase:

    --user USER       (See description above.)

    --root PATH       The location of the source tree.

    --dist PATH       Dist directory to use.  This is passed to buildenv.

    --path PATH       Directory where to run build command (default is --root).

  Optional when using --phase:

    --machine HOST    Machine name where to do the build.

    --command CMD     Command to run to build (default is buildenv).

    --checkdist SPEC  Fail is nothing under dist satisfies SPEC (relative).

    --clean_dist      Clean out dist root before starting.

    --sparse          Passed to buildenv.  Can use --nosparse.

  Required when using --machine:

    --ssh_command     (See description above.)

    --scp_command     (See description above.)

  Optional when using --machine:

    --scpdist         NEED DESCRIPTION (see .conf file format).

  Optional, passed directly to buildenv:

    --debug           Set up for debug builds (default).  Can use --nodebug.

    --optimize        Setup for optimized builds.  Can use --nooptimize.
DATA
}


########################################################################
#
# Copied from BuildUtils.pm since we cannot depend on it when
# we push this script over.
#
# Note that perl on Windows appears to have a bug where perl fails to pass all
# of the command-line.  We might want to have some sort of sanity-checking in
# Sytem and BackTick.
#

my @DIR_STACK;


sub PushDir($)
{
    my $path = shift || die;
    my $cwd = getcwd() || die;

    push(@DIR_STACK, $cwd);

    if ( $TEST_MODE )
    {
        print "CHDIR: $path\n";
        return;
    }

    chdir( $path ) || die "Could not chdir to \"$path\"\n";
}


sub PopDir()
{
    my $path = pop(@DIR_STACK) || die "Nothing on directory stack!\n";

    if ( $TEST_MODE )
    {
        print "CHDIR: $path\n";
        return;
    }

    chdir( $path ) || die "Could not chdir to \"$path\"\n";
}


sub GetErrorCodeFromChildError($)
{
    my $error = shift;

    if ( $error == -1 )
    {
        return $error;
    }
    else
    {
        return $error >> 8;
    }
}


sub System($)
{
    my $command = shift || die;

    system( $command." 2>&1");
    return GetErrorCodeFromChildError($?);
}


sub BackTick($$)
{
    my $command = shift || die;
    my $lines = shift;

    die if not ref($lines) eq 'ARRAY';

    @$lines = `$command 2>&1`;
    my $error = GetErrorCodeFromChildError($?);
    if ( $error )
    {
        return $error;
    }

    map {
        chomp($_);
    } @$lines;

    return 0;
}


sub RemoveLastPathComponent($)
{
    my $path = shift || die;

    my ($volume, $directories, $file) = File::Spec->splitpath( $path, 1 );

    die if $file;

    my @dirs = File::Spec->splitdir( $directories );
    pop(@dirs);

    if ( $volume ne '' )
    {
        unshift(@dirs, $volume);
    }

    return File::Spec->catfile(@dirs);
}


sub QuoteAsNeeded($)
{
    my $arg = shift || '';
    return ( $arg =~ / / ) ? "\"$arg\"" : $arg;
}


########################################################################
#
# The code in this section cannot depend on anything from any of
# the modules in the tree.
#

sub CreateDirectory($)
{
    my $path = shift || die;

    if ( $TEST_MODE )
    {
        print "[MKPATH: $path]\n";
        return;
    }

    mkpath("$path", 0, 0755);
}


sub FakeRun($)
{
    my $command = shift || die;

    print "FAKERUN: $command\n";
}


sub RunEx($$;$$)
{
    my $command = shift || die;
    my $print = shift;
    my $backtick = shift;
    my $noDie = shift;

    print "RUN: $command\n" if $print;

    if ( $TEST_MODE )
    {
        my $extraInfo = $print ? '' : ": $command";
        print '[TEST MODE RUN',$extraInfo,']',"\n";
        return 0;
    }

    my $error = $backtick ? BackTick( $command, $backtick ) : System( $command );
    if ( $error )
    {
        if ( $noDie )
        {
            return $error;
        }
        die "FAILED RUN - exit code $error for command below:\n".
            "    $command\n";
    }
}


sub Run($)
{
    my $command = shift || die;

    RunEx( $command, 1 );
}


sub RunNoDie($)
{
    my $command = shift || die;

    RunEx( $command, 1, 0, 1 );
}


sub RunBackTick($)
{
    my $command = shift || die;
    my @lines;

    RunEx( $command, 0, \@lines );

    return @lines;
}


sub BgDelTree($)
{
    my $path = shift || die;

    #
    # TODO -- Implement properly...
    #

    FakeRun("bgdeltree $path");
}


sub Scorch($)
{
    my $path = shift || die;

    #
    # This function removes any unknown files.  It assumes that a revert has
    # already happened (though we can change that).
    #

    foreach ( RunBackTick($SVN.' status --no-ignore ' . QuoteAsNeeded($path) ) )
    {
        if ( /^[?I] *(.*)$/ )
        {
            print "Removing: $1\n";
            rmtree( $1 ) || die "Unable to remove $1\n";
        }
    }
}


sub SvnList($$)
{
    my $path = shift || die;
    my $revision = shift || die;

    return RunBackTick($SVN.' ls -r '.QuoteAsNeeded($revision).' '.QuoteAsNeeded($path));
}


sub SvnListSubDirs($$)
{
    my $path = shift || die;
    my $revision = shift || die;

    return grep(/\/$/, SvnList( $path, $revision ));
}


sub SvnUpdate($$)
{
    my $path = shift || die;
    my $revision = shift || die;

    Run($SVN.' up -r '.QuoteAsNeeded($revision).' '.QuoteAsNeeded($path));
}


sub SvnUpdateForceSubDirs($$)
{
    my $path = shift || die;
    my $revision = shift || die;

    #
    # ISSUE - svn up does not have a --force-recurse option to pull
    # down the rest of everything.  You have to name the directories
    # to pull down.  So that's what we try to do.
    #
    # However, we only do it for the top level.  To do the full
    # recursion, we would have to recurse the tree and compare that to
    # a recursive svn list.
    #

    SvnUpdate( $path, $revision );

    my @dirs = SvnListSubDirs( $path, $revision );
    foreach my $dir (@dirs)
    {
        my $fullDir = File::Spec->catdir( $path, $dir );
        if ( ! -d $fullDir )
        {
            SvnUpdate( $fullDir, $revision );
        }
    }
}


sub SvnRevert($)
{
    my $path = shift || die;
    Run($SVN.' revert -R '.QuoteAsNeeded($path));
}


sub SvnCleanup($)
{
    my $path = shift || die;
    Run($SVN.' cleanup '.QuoteAsNeeded($path));
}


sub SvnCheckOut($$$)
{
    my $root = shift || die;
    my $url = shift || die;
    my $revision = shift || die;
    Run($SVN.' co -r '.QuoteAsNeeded($revision).' '.QuoteAsNeeded($url).' '.QuoteAsNeeded($root));
}


sub Trim($)
{
    my $s = shift;
    $s =~ s/^\s*//;
    $s =~ s/\s*$//;
    return $s;
}


sub EvaluateCondition($$)
{
    my $expression = shift;
    my $phase = shift || die;

    die if not defined $expression;

    if ( $expression eq '0' ||
         $expression eq '1' )
    {
        return $expression;
    }
    elsif ( $expression =~ /^\s*(\S+)\s*==\s*(\S+)\s*$/ )
    {
        return ($1 eq $2) ? 1 : 0;
    }
    elsif ( $expression =~ /^\s*(\S+)\s*!=\s*(\S+)\s*$/ )
    {
        return ( not ($1 eq $2) ) ? 1 : 0;
    }
    else
    {
        die "Cannot evaluate condition for phase $phase:\n$expression\n";
    }
}


sub EvaluateScpDist($$)
{
    my $expression = shift || die;
    my $phase = shift || die;

    if ( $expression =~ /^([^,]+),([^,]+)?,([^,]+)$/ )
    {
        return
        {
         local => $1,
         pre => $2,
         post => $3,
        };
    }
    else
    {
        die "Cannot evaluate scpdist for phase $phase:\n$expression\n";
    }
}


sub ReadConfigError($$$$)
{
    my $message = shift || die;
    my $file = shift || die;
    my $lineIndex = shift;
    my $line = shift;

    my $lineNumber = $lineIndex + 1;

    my $error = "$message in file \"$file\" line $lineIndex";
    if ( $line )
    {
        die "$error:\n$line\n";
    }
    else
    {
        die "$error\n"
    }
}

sub ListToHash($;$)
{
    my $list = shift || die;
    my $value = shift;

    die if not (ref($list) eq 'ARRAY');
    if (not defined $value)
    {
        $value = 1;
    }

    my $hash = {};
    map { $hash->{$_} = $value; } @$list;

    return $hash;
}

sub ExpandValue($$$;$)
{
    my $defs = shift || die;
    my $value = shift;
    my $key = shift;
    my $phase = shift;

    while ($value =~ /\$\{([^}]+)\}/)
    {
        if ( not exists $defs->{$1} )
        {
            my $error = "Referencing undefined variable $1 from variable $key";
            if ($phase)
            {
                die "$error in phase $phase.\n";
            }
            else
            {
                die "$error.\n";
            }
        }
        $value = $`.$defs->{$1}.$';
    }
    return $value;
}

sub InternalReadConfig($)
{
    my $file = shift || die;

    my $fh = new IO::File "<$file";
    die "Unable to open $file" if not defined $fh;

    my $phaseNames = [];
    my $phaseIndex = {};
    my $config = {};
    my $section = undef;
    my $isListMode = 0;
    my $validKeys = undef;

    my @lines = $fh->getlines();
    chomp(@lines);
    for (my $i = 0; $i <= $#lines; $i++)
    {
        my $line = Trim($lines[$i]);
        $line =~ s/^#.*//;

        if ($line eq '')
        {
            next;
        }
        elsif ($line =~ /^\[(.*)\]$/ )
        {
            $section = Trim($1);

            my $phase = undef;
            $isListMode = 0;
            $validKeys = undef;

            if ($section eq '')
            {
                ReadConfigError("Empty section name",
                                $file, $i, $line);
            }
            elsif ($section eq 'phases')
            {
                $isListMode = 1;
            }
            elsif ($section eq 'vars')
            {
                # ok
            }
            elsif ($section eq 'options')
            {
                $validKeys = ListToHash(\@OPTIONAL_OPTIONS_KEYS);
            }
            elsif ($section =~ /^phases\/($ID_PATTERN)/)
            {
                $phase = $1;
                $validKeys = ListToHash(\@VALID_PHASE_KEYS);
            }
            else
            {
                ReadConfigError("Unrecognized section name \"$section\"",
                                $file, $i, $line);
            }

            if (exists $config->{$section})
            {
                ReadConfigError("Duplicate section name \"$section\"",
                                $file, $i, $line);
            }

            $config->{$section}->{_list} = [];
            $config->{$section}->{_hash} = {};

            if ( defined $phase )
            {
                $phaseIndex->{$phase} = $config->{$section}->{_hash};
                $phaseIndex->{$phase}->{name} = $phase;
                push(@$phaseNames, $phase);
            }

            next;
        }
        else
        {
            if ( not defined $section )
            {
                ReadConfigError("Not inside any section",
                                $file, $i, $line);
            }
            elsif ( $line =~ /^($ID_PATTERN)\s*=\s*(.*)$/ )
            {
                my $key = $1;
                my $value = $2;

                if ( $isListMode )
                {
                    ReadConfigError("Key/value pair is not valid in section \"$section\"",
                                    $file, $i, $line);
                }
                if ( exists $config->{$section}->{_hash}->{$key} )
                {
                    ReadConfigError("Key \"$key\" already exists in section \"$section\"",
                                    $file, $i, $line);
                }
                if ( $validKeys and (not exists $validKeys->{$key}) )
                {
                    map { print "$_=$validKeys->{$_}\n"; } keys %$validKeys;
                    ReadConfigError("Unknown key \"$key\" in section \"$section\"",
                                    $file, $i, $line);
                }
                push(@{$config->{$section}->{_list}}, $key);
                $config->{$section}->{_hash}->{$key} = $value;
                next;
            }
            elsif ( $line =~ /^($ID_PATTERN)$/ )
            {
                my $key = $1;

                if ( not $isListMode )
                {
                    ReadConfigError("Invalid key/value directive in section \"$section\"",
                                    $file, $i, $line);
                }
                if ( exists $config->{$section}->{_hash}->{$key} )
                {
                    ReadConfigError("Item \"$key\" already exists in section \"$section\"",
                                    $file, $i, $line);
                }
                push(@{$config->{$section}->{_list}}, $key);
                $config->{$section}->{_hash}->{$key} = 1;
                next;
            }
            else
            {
                ReadConfigError("Invalid directive in section \"$section\"",
                                $file, $i, $line);
            }
            die "unreachable code";
        }
        die "unreachable code";
    }

    return { config => $config, phaseIndex => $phaseIndex, phaseNames => $phaseNames };
}


sub ReadConfig($$$)
{
    my $file = shift || die;
    my $argOpts = shift || die;
    my $defineOpts = shift || die;

    my $parseInfo = InternalReadConfig($file);

    my $phaseNames = $parseInfo->{phaseNames} || die;
    my $phaseIndex = $parseInfo->{phaseIndex} || die;
    my $config = $parseInfo->{config} || die;
    my $defs = {};

    #
    # First, do the stuff that comes as direct args
    #

    foreach my $key (@REQUIRED_OPTIONS_KEYS)
    {
        if (not exists $argOpts->{$key})
        {
            # should never happen, so no terminating \n
            die "Missing required options key \"$key\" from args.";
        }
        $defs->{$key} = $argOpts->{$key};
    }

    foreach my $key (@OPTIONAL_OPTIONS_KEYS)
    {
        if ( exists $argOpts->{$key} )
        {
            if ( exists $config->{options}->{_hash}->{$key} )
            {
                print "Command-line overrides [options] for: $key.\n";
            }
            $defs->{$key} = $argOpts->{$key};
        }
    }

    foreach my $key (@VALID_OPTIONS_KEYS)
    {
        if ( exists $defineOpts->{$key} )
        {
            die "Cannot redefine an options key: $key\n";
        }
    }

    #
    # Then, do user defines.
    #

    foreach my $key (keys %$defineOpts)
    {
        $defs->{$key} = $defineOpts->{$key};
    }

    #
    # Now evaluate [options] using user defines and args.
    #

    foreach my $key (@{$config->{options}->{_list}})
    {
        if (exists $defs->{$key})
        {
            #print "Command line overrides [options] for: $key.\n";
        }
        else
        {
            my $value = $config->{options}->{_hash}->{$key};
            $value = ExpandValue($defs, $value, $key);
            $defs->{$key} = $value;
        }
    }

    #
    # Define OPTIONAL_OPTIONS_KEYS if not already defined.
    #

    if ( not $defs->{user} )
    {
        $defs->{user} = $ENV{USER} || $ENV{USERNAME} || die "Could not determine user.\n";
    }

    if ( not $defs->{scp_command} )
    {
        $defs->{scp_command} = 'scp';
    }

    if ( not $defs->{ssh_command} )
    {
        $defs->{ssh_command} = 'ssh';
    }

    #
    # Evaluate regular variables [vars].
    #

    foreach my $key (@{$config->{vars}->{_list}})
    {
        my $value = $config->{vars}->{_hash}->{$key};
        $value = ExpandValue($defs, $value, $key);
        if ( exists $defs->{$key} )
        {
            print "Ignoring already defined var: $key.\n";
        }
        else
        {
            $defs->{$key} = $value;
        }
    }

    #
    # Build up lists of phases we want to try running.
    #

    my $phaseList = [];
    my $phaseListHash = {};
    foreach my $phaseName (@{$config->{phases}->{_list}})
    {
        # Do we really want to disallow this?
        if ( exists $phaseListHash->{$phaseName} )
        {
            die "Phase $phaseName is multiply listed in [phases] section\n";
        }
        push(@$phaseList, $phaseName);
        $phaseListHash->{$phaseName} = 1;
    }

    if ($#{$phaseList} < 0)
    {
        die "No phases in [phases] section!\n"
    }

    #
    # Expand all phases and verify keys for phases in list.
    #

    foreach my $phaseName (@$phaseNames)
    {
        if (not exists $phaseIndex->{$phaseName})
        {
            die "Unknown phase \"$phaseName\" in [phases] section.\n";
        }

        my $phaseInfo = $phaseIndex->{$phaseName};
        if ( exists $phaseInfo->{depends} )
        {
            # if we inherited, we would have expanded into an array already
            if (not ref($phaseInfo->{depends}))
            {
                my @depNames = split(/[,\s]+/, $phaseInfo->{depends});
                if ($#depNames < 0)
                {
                    delete $phaseInfo->{depends};
                }
                else
                {
                    my $deps = [];
                    foreach my $depName (@depNames)
                    {
                        if (not exists $phaseIndex->{$depName})
                        {
                            die "Phase \"$phaseName\" depends on non-existent phase \"$depName\".\n";
                        }
                        push(@$deps, $phaseIndex->{$depName});
                    }
                    $phaseInfo->{depends} = $deps;
                }
            }
        }
        if (exists $phaseInfo->{inherit})
        {
            my $inheritName = $phaseInfo->{inherit};
            if (not exists $phaseIndex->{$inheritName})
            {
                die "Cannot inherit from non-existent phase \"$phaseInfo->{inherit}\" in $phaseName.\n";
            }
            my $inheritInfo = $phaseIndex->{$inheritName};
            foreach my $key (keys %{$inheritInfo})
            {
                if (exists $phaseInfo->{$key})
                {
                    # For depends, we need to add the inheritance.  Note
                    # that order does not matter for depends.  However,
                    # we try to preserve the inherited one first to make
                    # the execution order more predictable from looking
                    # at the config file.
                    if ($key eq 'depends')
                    {
                        $phaseInfo->{$key} = [ @{$inheritInfo->{$key}}, @{$phaseInfo->{$key}} ];
                    }
                    print "Not inheriting $key from $inheritName in $phaseName.\n";
                }
                else
                {
                    $phaseInfo->{$key} = $inheritInfo->{$key};
                }
            }
        }

        #
        # Verify required keys for phases in run list.
        #

        if ($phaseListHash->{$phaseName})
        {
            foreach my $key (@REQUIRED_PHASE_KEYS)
            {
                if (not exists $phaseInfo->{$key})
                {
                    die "Missing required key \"$key\" in phase \"$phaseName\".\n";
                }
            }
        }

        foreach my $key (keys %$phaseInfo)
        {
            my $value = $phaseInfo->{$key};
            $phaseInfo->{$key} = ExpandValue($defs, $value, $key, $phaseName);
        }
        if ( exists $phaseInfo->{cond} )
        {
            $phaseInfo->{cond} = EvaluateCondition($phaseInfo->{cond}, $phaseName);
        }
        if ( exists $phaseInfo->{scpdist} )
        {
            $phaseInfo->{scpdist_orig} = $phaseInfo->{scpdist};
            $phaseInfo->{scpdist} = EvaluateScpDist($phaseInfo->{scpdist}, $phaseName);
        }
    }

    return
    {
     phases =>
     {
      index => $phaseIndex,
      list => $phaseList,
      names => $phaseNames,
     },
     opts =>
     {
      user => $defs->{user},
      scp_command => $defs->{scp_command},
      ssh_command => $defs->{ssh_command},
     },
    };
}


sub PrepareTree($$$;$)
{
    my $BuildRepository = shift || die;
    my $BuildRevision = shift || die;
    my $BuildRoot = shift || die;
    my $doRevert = shift || 0;

    #
    # If the tree does not exist, check it all out.
    #

    if ( ! -d $BuildRoot )
    {
        SvnCheckOut( $BuildRoot, $BuildRepository, $BuildRevision );
    }
    else
    {
        #
        # Otherwise, update it.  Note that it could have been the case that
        # someone did a non-recursive check-out.  (This might happen if
        # someone had to re-initialize the tree on a build machine for some
        # build software.  In such a case, it may not be desirable to wait for
        # a full checkout before starting the build software.)  Therefore, we
        # do a forceful update of the (top-level) subdirectories just in such
        # a case.
        #

        eval {
            SvnUpdateForceSubDirs( $BuildRoot, $BuildRevision );
        };
        if ( $@ )
        {
            print "Unable to update subdirs. Cleaning up and trying one more time.\n";
            SvnCleanup( $BuildRoot );
            SvnUpdateForceSubDirs( $BuildRoot, $BuildRevision );
        }
        if ( $doRevert )
        {
            SvnRevert( $BuildRoot );
            Scorch( $BuildRoot );
        }
    }
}


sub IsWindows()
{
    my $osName = $Config{osname} || die;

    if ( ($osName eq 'MSWin32') or
         ($osName eq 'MSWin64') or
         ($osName eq 'cygwin') )
    {
        return 1;
    }

    return 0;
}

sub NoPrepareTree($)
{
    my $BuildRoot = shift || die;

    if (! -d $BuildRoot)
    {
        die "Missing build root when not preparing: '$BuildRoot'\n";
    }

    if (!IsWindows())
    {
        Run("while true; do time find $BuildRoot/. -perm 0000 && break; done");
    }
}


########################################################################
#
# The code in this section can depend on modules from the tree.
#

sub CreateArgsLine($$)
{
    my $opts = shift || die;
    my $map = shift || die;

    my $argsLine = '';
    foreach my $optionName (keys %$opts)
    {
        next if not defined($map->{$optionName});

        my $optionType = $map->{$optionName};
        my $optionNewName = $optionName;

        if ($optionType =~ /^([^:]*):(.*)$/)
        {
            $optionType = $1;
            $optionNewName = $2;
        }

        if ( $optionType eq '' )
        {
            if ( $opts->{$optionName} )
            {
                $argsLine .= ' --' . $optionNewName;
            }
        }
        elsif ( $optionType eq 's' )
        {
            $argsLine .= ' --' . $optionNewName . ' '. QuoteAsNeeded($opts->{$optionName});
        }
        elsif ( $optionType eq '!' )
        {
            $argsLine .= ' --' . ($opts->{$optionName} ? '' : 'no') . $optionNewName;
        }
        else
        {
            die;
        }
    }

    return $argsLine;
}


sub CreateOfficialBuildArgsLine($$;$)
{
    my $opts = shift || die;
    my $phaseOpts = shift || die;
    my $do_remote = shift || 0;

    if ($do_remote and IsPhaseLocal($phaseOpts))
    {
        $do_remote = 0;
    }

    my @argsLines = ();

    my $map =
    {
     revert => '',

     name => 's',
     number => 's',
     repository => 's',
     revision => 's',
     user => 's',

     debug => '!',
     optimize => '!',
    };

    push(@argsLines, CreateArgsLine( $opts, $map ));

    my $phaseMap =
    {
     name => 's:phase',
     root => 's',
     dist => 's',
     image => 's',
     meta => 's',
     run => 's',
     cache => 's',
     path => 's',
     command => 's',
     checkdist => 's',
     sparse => '!',
     noprepare => '',
     nobuildenv => '',
    };

    push(@argsLines, CreateArgsLine( $phaseOpts, $phaseMap ));

    if ($do_remote)
    {
        my $remoteMap =
        {
         ssh_command => 's',
         scp_command => 's',
        };

        push(@argsLines, CreateArgsLine( $opts, $remoteMap ));

        my $remotePhaseMap =
        {
         machine => 's',
         scpdist_orig => 's:scpdist',
        };

        push(@argsLines, CreateArgsLine( $phaseOpts, $remotePhaseMap ));
    }

    return join(' ', @argsLines);
}


sub CreateBuildEnvArgsLine($)
{
    my $opts = shift || die;

    #
    # Note that we can re-map the options here if the argument names for
    # buildenv are different than the argument names for official-build.
    #

    my $map =
    {
     phase => 's:official_build_phase',
     name => 's:official_build_name',
     number => 's:official_build_number',
     repository => 's:official_build_sc_repository',
     revision => 's:official_build_sc_revision',
     user => 's',
     dist => 's',
     image => 's',
     meta => 's',
     cache => 's',
     run => 's',
     debug => '!',
     optimize => '!',
     sparse => '',
    };

    return CreateArgsLine( $opts, $map );
}


sub GetRandomFileChar()
{
    my $alphaChars = 'abcdefghijklmnopqrstuvwxyz';
    my $numChars = '0123456789';
    my $otherChars = '_';

    my $allowedChars = $alphaChars . uc($alphaChars) . $numChars . $otherChars;
    return substr($allowedChars, int(rand(length($allowedChars))), 1);
}


sub GetRandomFileChars($)
{
    my $len = shift || 0;

    my $result = '';
    for (my $i = 0; $i < $len; $i++)
    {
        $result .= GetRandomFileChar();
    }

    return $result;
}


sub TestSsh($$)
{
    my $sshCommand = shift || die;
    my $machine = shift || die;
    my $port = 22;
    my $sshPortOption = "-P";
    if ( $machine =~ /^(.*):([0-9]+)$/ )
    {
        $port = $2;
        $machine = $1;
    }

    print "TESTING: ssh to $machine\n";
    RunEx( "$sshCommand $sshPortOption $port $machine echo SUCCESS: connection works", 0);
}


sub DoScpDist($$$)
{
    my $opts = shift || die;
    my $phaseInfo = shift || die;
    my $part = shift || die;

    my $scpCommand = $opts->{scp_command} || die;
    my $sshCommand = $opts->{ssh_command} || die;
    my $sshPortOption = "-P";
    my $scpPortOption = "-P";

    my $machine = $phaseInfo->{machine} || die;
    my $port = 22;
    if ( $machine =~ /^(.*):([0-9]+)$/ )
    {
        $port = $2;
        $machine = $1;
    }
    my $dist = $phaseInfo->{dist} || die;
    my $scpDist = $phaseInfo->{scpdist};

    if ( $scpDist )
    {
        my $localDist = $scpDist->{local} || die;
        my $relPath = $scpDist->{$part};

        if ( ($part eq 'pre') and (not $relPath) )
        {
            return;
        }

        die if not $relPath;

        my $remote = "$dist/$relPath";
        my $local = "$localDist/$relPath";

        my $from;
        my $to;

        if ( $part eq 'pre')
        {
            # Run("$sshCommand $machine perl -MFile::Path -e \"\\"mkpath($to, 0, 0755);\\"\"");
            Run("$sshCommand $sshPortOption $port $machine mkdir -p $remote");
            $from = "$local";
            $to = "$machine:$remote";
        }
        elsif ( $part eq 'post' )
        {
            CreateDirectory($local);
            $from = "$machine:$remote";
            $to = "$local";
        }
        else
        {
            die "Invalid part: $part";
        }

        die if not $from;
        die if not $to;

        Run("$scpCommand $scpPortOption $port -r -p $from/* $to/.");
    }
}


sub DoCleanScpDist($$)
{
    my $opts = shift || die;
    my $phaseInfo = shift || die;

    my $scpCommand = $opts->{scp_command} || die;
    my $sshCommand = $opts->{ssh_command} || die;
    my $sshPortOption = "-P";
    my $scpPortOption = "-P";

    my $machine = $phaseInfo->{machine} || die;
    my $port = 22;
    if ( $machine =~ /^(.*):([0-9]+)$/ )
    {
        $port = $2;
        $machine = $1;
    }
    my $dist = $phaseInfo->{dist} || die;
    my $scpDist = $phaseInfo->{scpdist};

    if ( $scpDist )
    {
        # If we had to copy dist, it means that the dist is not shared, so we
        # will delete it.

        # Before deleting it, we want to make super-sure that it is
        # not some important directory: (/, /etc/*, /lib{32,64}/*,
        # /etc/*, /var/{lib,cache,log,adm}/*.

        $dist =~ s/^\s+//;
        $dist =~ s/\s+$//;
        if (($dist =~ /^\/+$/) or
            ($dist =~ /^\/+usr($|\/+)/) or
            ($dist =~ /^\/+lib(32|64)?($|\/+)/) or
            ($dist =~ /^\/+etc($|\/+)/) or
            ($dist =~ /^\/+var\/+lib($|\/+)/) or
            ($dist =~ /^\/+var\/+log($|\/+)/) or
            ($dist =~ /^\/+var\/+adm($|\/+)/) or
            ($dist =~ /^\/+var\/+cache($|\/+)/))
        {
            print "Not deleting '$dist' since it matches a protected prefix.\n";
        }
        else
        {
            Run("$sshCommand $sshPortOption $port $machine rm -rf $dist");
        }
    }
}


sub BuildRemote($$)
{
    my $opts = shift || die;
    my $phaseInfo = shift || die;

    my $doRevert = $opts->{revert} || 0;

    my $name = $opts->{name} || die;
    my $repository = $opts->{repository} || die;
    my $revision = $opts->{revision} || die;

    my $scpCommand = $opts->{scp_command} || die;
    my $sshCommand = $opts->{ssh_command} || die;
    my $sshPortOption = "-P";
    my $scpPortOption = "-P";

    my $machine = $phaseInfo->{machine} || die;
    my $port = 22;
    if ( $machine =~ /^(.*):([0-9]+)$/ )
    {
        $port = $2;
        $machine = $1;
    }
    my $root = $phaseInfo->{root} || die;
    my $dist = $phaseInfo->{dist} || die;
    my $scpDist = $phaseInfo->{scpdist} || 0;

    my $filePart = fileparse($0);
    my $file = File::Spec->catfile( $FindBin::Bin, $filePart );

    my $argsLine = CreateOfficialBuildArgsLine($opts, $phaseInfo);

    print "Preparing for remote build on $machine\n";
    TestSsh($sshCommand, $phaseInfo->{machine});

    my $failed = 0;
    my $tempFile = "/tmp/$filePart.$name.$revision.".GetRandomFileChars(10);
    Run("$scpCommand $scpPortOption $port ".QuoteAsNeeded($file)." ".QuoteAsNeeded("$machine:$tempFile"));
    eval {
        Run("$sshCommand $sshPortOption $port $machine chmod a+x ".QuoteAsNeeded($tempFile));
        DoScpDist($opts, $phaseInfo, 'pre');
        Run("$sshCommand $sshPortOption $port $machine perl ".QuoteAsNeeded($tempFile)." $argsLine");
        DoScpDist($opts, $phaseInfo, 'post');
    };
    if ( $@ )
    {
        print STDERR $@;
        $failed = 1;
    }
    Run("$sshCommand $sshPortOption $port $machine rm -f ".QuoteAsNeeded($tempFile));
    DoCleanScpDist($opts, $phaseInfo);
    if ( $failed )
    {
        exit(1);
    }
}


sub CreateBuildEnvInvocation($$)
{
    my $root = shift || die;
    my $argsLine = shift || '';

    my $buildenv = File::Spec->catfile( File::Spec->catdir( $root, 'build'), 'buildenv' );

    if ( IsWindows() )
    {
        my $osName = $Config{osname} || die;
        if ($osName eq 'cygwin')
        {
            my @lines = RunBackTick( "cygpath -w \"$buildenv\"" );
            $buildenv = $lines[0];
        }
        return "cmd /c \"".QuoteAsNeeded("$buildenv.cmd")." $argsLine\"";
    }
    else
    {
        my $osName = $Config{osname} || die;
        my $extraOpts = "";

        #
        # Escape backslashes and quote in argument line.  Note that we
        # must do the additional quoting when calling bash built-ins (such
        # as source).  Therefore, we must escape the quotes we already
        # have.
        #

        my $fullArgsLine = QuoteAsNeeded($buildenv).' '.$argsLine;
        $fullArgsLine =~ s/\\/\\\\/g;
        $fullArgsLine =~ s/"/\\"/g;

        if ($osName eq 'hpux' || $osName eq 'darwin')
	{
	    $extraOpts = "-l";
	}

	return "bash $extraOpts -c \"source $fullArgsLine\"";
    }
}


sub CreateCommandInvocation($$)
{
    my $command = shift || die;
    my $opts = shift || die;

    return "$command";
}

sub BuildLocal($)
{
    my $opts = shift || die;

    my $cleanDist = $opts->{clean_dist} || 0;
    my $dist = $opts->{dist} || die;
    my $root = $opts->{root} || die;
    my $path = $opts->{path};
    my $noprepare = $opts->{noprepare};
    my $nobuildenv = $opts->{nobuildenv};
    my $command = $opts->{command} || 'buildtool';
    my $checkdist = $opts->{checkdist};

    if ( !$noprepare )
    {
        PrepareTree( $opts->{repository}, $opts->{revision}, $root, $opts->{revert} );
    }
    else
    {
        NoPrepareTree($root);
    }

    if ( $cleanDist )
    {
        BgDelTree( $dist ); # or fail if exists?
    }

    PushDir( $root );
    PushDir( $path ) if $path;
    if ($nobuildenv)
    {
        Run(CreateCommandInvocation($command, $opts));
    }
    else
    {
        $ENV{OFFICIAL_BUILD_RANDOM} = int(rand(1000000));
        my $argsLine = CreateBuildEnvArgsLine( $opts );
        Run(CreateBuildEnvInvocation($root, "$argsLine --exec buildtool clean"));
        Run(CreateBuildEnvInvocation($root, "$argsLine --exec $command"));
    }
    PopDir() if $path;
    PopDir();

    if ( $checkdist )
    {
        my @files = glob(File::Spec->catfile($dist, $checkdist));
        if ( $#files < 0 )
        {
            die "ERROR: checkdist detected \"$checkdist\" missing from dist.\n";
        }
        else
        {
            print "INFO: checkdist found at least: \"$files[0]\"\n";
        }
    }

    #
    # TODO -- copy log files? (would need log file location, perhaps?)
    #
}


sub IsPhaseLocal($)
{
    my $phaseInfo = shift || die;

    return ( $phaseInfo->{machine} eq '.' ) ? 1 : 0;
}

sub IsPhaseEnabled($)
{
    my $phaseInfo = shift || die;

    if ( ( exists $phaseInfo->{cond} ) and
         ( not $phaseInfo->{cond} ) )
    {
        return 0;
    }
    return 1;
}


sub PhaseSetStatus($$;$)
{
    my $phaseInfo = shift || die;
    my $status = shift || die;
    my $subStatus = shift;

    $phaseInfo->{_}->{status} = $status;
    if ($subStatus)
    {
        $phaseInfo->{_}->{subStatus} = $subStatus;
    }
}


sub PhaseGetStatus($)
{
    my $phaseInfo = shift || die;

    if (exists $phaseInfo->{_}->{status})
    {
        return $phaseInfo->{_}->{status};
    }

    return undef;
}


sub PhaseGetSubStatus($)
{
    my $phaseInfo = shift || die;

    if (exists $phaseInfo->{_}->{subStatus})
    {
        return $phaseInfo->{_}->{subStatus};
    }

    return undef;
}


sub PhaseSetPid($$)
{
    my $phaseInfo = shift || die;
    my $pid = shift;

    if ( not $pid )
    {
        delete $phaseInfo->{_}->{pid};
    }
    else
    {
        $phaseInfo->{_}->{pid} = $pid;
    }
}


sub PhaseGetPid($)
{
    my $phaseInfo = shift || die;

    if (exists $phaseInfo->{_}->{pid})
    {
        return $phaseInfo->{_}->{pid};
    }

    return undef;
}


sub PhaseTimerStart($)
{
    my $phaseInfo = shift || die;

    $phaseInfo->{_}->{time}->{start} = time();
}


sub PhaseTimerStop($)
{
    my $phaseInfo = shift || die;

    my $now = time();

    die if not exists $phaseInfo->{_}->{time}->{start};

    my $start = $phaseInfo->{_}->{time}->{start};
    $phaseInfo->{_}->{time}->{elapsed} = $now - $start;
    delete $phaseInfo->{_}->{time}->{start};
}


sub PhaseTimerGetElapsed($)
{
    my $phaseInfo = shift || die;

    if (exists $phaseInfo->{_}->{time}->{elapsed})
    {
        return $phaseInfo->{_}->{time}->{elapsed};
    }

    return undef;
}


sub CheckSkip($)
{
    my $phases = shift || die;

    my $phaseList = $phases->{list} || die;
    my $phaseIndex = $phases->{index} || die;
    my $phaseNames = $phases->{names} || die;

    my $inList = {};

    print "Checking for phases to skip.\n";

    foreach my $phaseName (@$phaseList)
    {
        $inList->{$phaseName} = 1;
    }

    foreach my $phaseName (@$phaseNames)
    {
        my $phaseInfo = $phaseIndex->{$phaseName} || die;

        if ( not IsPhaseEnabled($phaseInfo) )
        {
            print "Skipping phase $phaseName as it does not meet conditions.\n";
            PhaseSetStatus($phaseInfo, $STATUS_SKIPPED);
        }
        elsif ( not $inList->{$phaseName} )
        {
            PhaseSetStatus($phaseInfo, $STATUS_SKIPPED);
        }
    }

    print "Checking dependencies.\n";

    #
    # In check dependencies, we rely on the fact that phases can only
    # depend on previous phases.
    #

    foreach my $phaseName (@$phaseList)
    {
        my $phaseInfo = $phaseIndex->{$phaseName} || die;
        if ( not PhaseGetStatus($phaseInfo) and $phaseInfo->{depends} )
        {
            foreach my $depInfo (@{$phaseInfo->{depends}})
            {
                my $status = PhaseGetStatus($depInfo);
                if ($status and ($status eq $STATUS_SKIPPED))
                {
                    die "Phase $phaseName depends on a phase that is not being built: $depInfo->{name}\n";
                }
            }
        }
    }
}


sub CheckConnections($$)
{
    my $sshCommand = shift || die;
    my $phases = shift || die;

    my $phaseList = $phases->{list} || die;
    my $phaseIndex = $phases->{index} || die;

    print "Checking connections.\n";

    foreach my $phaseName (@$phaseList)
    {
        my $phaseInfo = $phaseIndex->{$phaseName} || die;

        if ( IsPhaseEnabled($phaseInfo) )
        {
            if ( IsPhaseLocal($phaseInfo) )
            {
                print "Not checking connection for local phase $phaseName.\n";
                if ($phaseInfo->{depends})
                {
                    my @depNames;
                    map { push(@depNames, $_->{name}); } @{$phaseInfo->{depends}};
                    print "$phaseName: ".join(":", @depNames)."\n";
                }
            }
            else
            {
                print "Checking connection for remote phase $phaseName...\n";
                TestSsh($sshCommand, $phaseInfo->{machine});
            }
        }
    }
}


sub MergeOptsIntoBuildOpts($$)
{
    my $opts = shift || die;
    my $phaseInfo = shift || die;

    my $phaseName = $phaseInfo->{name} || die;

    if ( !IsPhaseLocal($phaseInfo) )
    {
        die "Cannot merge opts into build opts because \"$phaseName\" is not a local phase.\n"
    }

    my $buildOpts = { phase => $phaseName };
    foreach my $key qw(root dist sparse path command checkdist)
    {
        if (exists $phaseInfo->{$key})
        {
            $buildOpts->{$key} = $phaseInfo->{$key};
        }
    }
    foreach my $key (@REQUIRED_OPTIONS_KEYS, qw(debug nodebug optimize nooptimize revert))
    {
        if (exists $opts->{$key})
        {
            $buildOpts->{$key} = $opts->{$key};
        }
    }
    return $buildOpts;
}


sub BuildPhase($$)
{
    my $opts = shift || die;
    my $phaseInfo = shift || die;

    die if PhaseGetStatus($phaseInfo);
    die if not IsPhaseEnabled($phaseInfo);

    my $phaseName = $phaseInfo->{name};

    PhaseTimerStart($phaseInfo);
    eval {
        if ( IsPhaseLocal($phaseInfo) )
        {
            print "Building local phase $phaseName\n";
            BuildLocal(MergeOptsIntoBuildOpts($opts, $phaseInfo));
        }
        else
        {
            print "Building remote phase $phaseName\n";
            BuildRemote($opts, $phaseInfo);
        }
    };
    PhaseTimerStop($phaseInfo);
    if ($@)
    {
        PhaseSetStatus($phaseInfo, $STATUS_ERROR);
        die $@;
    }
    else
    {
        PhaseSetStatus($phaseInfo, $STATUS_DONE);
    }
}


sub GetLogFileName($$;$)
{
    my $name = shift;
    my $template = shift || die;
    my $checkDir = shift || 0;

    die if not defined($name) and not $checkDir;

    my ($filePart, $dirPart, $extPart) = fileparse($template, '\\.[^\\.]+$');
    if ($checkDir)
    {
        if (not -d $dirPart)
        {
            die "Log directory \"$dirPart\"not found\n";
        }
        return undef;
    }

    if (not defined($extPart))
    {
        $extPart = '';
    }

    my $suffix = $name ? "-phase-$name" : '-results';
    return File::Spec->catfile($dirPart, $filePart.$suffix.$extPart);
}


sub SpawnRun($$)
{
    my $command = shift || die;
    my $file = shift || die;

    # This is the only way we figured out to hack this.
    # Appently, open3 only accepts bareword magic for
    # the redirect to an existing file descriptor.
    # It does not accept numbers, scalars, etc.
    # Note that we always kill off STDIN...

    local *FHO = new IO::File(">$file") || die "Could not open $file: $!\n";
    *FHO if 0; # to prevent warning about var not being used.
    use IPC::Open3;
    print "SPAWN: $command\n";
    my $pid = open3(">/dev/null", ">&FHO", ">&FHO", $command);
    if (not $pid)
    {
        die "Could not start: $command\n";
    }
    return $pid;
}


sub SpawnBuildPhase($$$)
{
    my $opts = shift || die;
    my $phaseInfo = shift || die;
    my $children = shift || die;

    my $machine = $phaseInfo->{machine} || die;

    if ($children->{machines}->{$machine})
    {
        return 0;
    }

    PhaseTimerStart($phaseInfo);

    print "\nStarting child phase $phaseInfo->{name}\n";

    my $file = GetLogFileName($phaseInfo->{name}, $opts->{log});
    print "Logging phase $phaseInfo->{name} to $file\n";

    my $argsLine = CreateOfficialBuildArgsLine($opts, $phaseInfo, 1);
    my $pid = SpawnRun("perl $0 $argsLine", $file);

    $children->{machines}->{$machine}++;
    $children->{pids}->{$pid} = $phaseInfo;

    PhaseSetPid($phaseInfo, $pid);

    $children->{count}++;

    print "Spawned child phase $phaseInfo->{name} @ $machine (pid = $pid, machine children = $children->{machines}->{$machine}, total children = $children->{count})\n";

    return 1;
}


sub WaitForChild($)
{
    my $children = shift || die;

    if ($children->{count} <= 0)
    {
        return 0;
    }

    print "\nWaiting for a child to finish.\n";

    my $pid = wait();
    my $error = GetErrorCodeFromChildError($?);

    my $phaseInfo = $children->{pids}->{$pid} || die "Bad pid: $pid";
    my $machine = $phaseInfo->{machine} || die;

    delete $children->{pids}->{$pid};

    PhaseSetPid($phaseInfo, 0);
    PhaseTimerStop($phaseInfo);
    if ($error)
    {
        PhaseSetStatus($phaseInfo, $STATUS_ERROR, $error);
    }
    else
    {
        PhaseSetStatus($phaseInfo, $STATUS_DONE);
    }

    $children->{machines}->{$machine}--;
    $children->{count}--;

    my $status = PhaseGetStatus($phaseInfo);
    print "Joined child phase $phaseInfo->{name} @ $machine (pid = $pid, machine children = $children->{machines}->{$machine}, total children = $children->{count}) --> $status\n";

    return 1;
}


sub ReadTailN($$;$)
{
    my $file = shift || die;
    my $numLines = shift || die;

    my $fh = new IO::File "<$file";
    die "Unable to open $file" if not defined $fh;

    $fh->binmode();
    $fh->sysseek(0, 2) || die "Unable to seek to end of $file\n";

    my $lineCount = 0;
    while ($lineCount <= $numLines)
    {
        $fh->sysseek(-1, 1) || last;
        $fh->sysread($_, 1, 0) || die "Failed to read $file\n";

        if ($_ eq "\n")
        {
            $lineCount++;
        }

        if ($lineCount > $numLines)
        {
            last;
        }

        $fh->sysseek(-1, 1) || last;
    }
    my $pos = $fh->sysseek(0, 1);
    die if not defined($pos);
    $pos = 0 if $pos eq '0 but true';

    $fh->seek($pos, 0) || die;
    my @lines = <$fh>;
    close $fh;

    map {
        $_ =~ s/\r|\n//g;
    } @lines;

    return @lines;
}


sub GetElapsedTimeString($)
{
    my $seconds = shift || 0;

    my $hours = int($seconds / 3600);
    $seconds -= $hours * 3600;

    my $minutes = int($seconds / 60);
    $seconds -= $minutes * 60;

    return $hours.'h'.$minutes.'m'.$seconds.'s';
}


# prototype for recursion
sub PrintResults($;$$);
sub PrintResults($;$$)
{
    my $phases = shift || die;
    my $opts = shift;
    my $logToFile = shift;

    my $errors = 0;

    my $phaseList = $phases->{list} || die;
    my $phaseIndex = $phases->{index} || die;

    my $haveLog = ($opts and $opts->{log}) ? 1 : 0;

    my $fh;
    if (not $logToFile)
    {
        $fh = \*STDERR;
    }
    else
    {
        my $file = $logToFile;
        $fh = new IO::File ">$file";
        die "Unable to open $file" if not defined $fh;
    }

    print $fh ("-" x 76)."\n";
    print $fh "Official Build Results\n";
    print $fh ("-" x 76)."\n";

    foreach my $phaseName (@$phaseList)
    {
        my $phaseInfo = $phaseIndex->{$phaseName} || die;
        my $status = PhaseGetStatus($phaseInfo) || 'N/A';
        my $subStatus = PhaseGetSubStatus($phaseInfo) || '';
        my $extra = $subStatus ? " $subStatus" : '';
        my $machine = $phaseInfo->{machine};
        my $time = PhaseTimerGetElapsed($phaseInfo);

        my $prettyTime = "";

        if (defined $time)
        {
            $prettyTime = ' '.GetElapsedTimeString($time);
        }

        if ($status eq $STATUS_ERROR)
        {
            $errors += 1;
        }

        print $fh "Phase $phaseName $status$extra $machine$prettyTime\n";

        if ($haveLog and ($status eq $STATUS_ERROR) and not ($subStatus eq $SUB_STATUS_DEP))
        {
            my $prefix = ' ' x 4;

            my $file = GetLogFileName($phaseInfo->{name}, $opts->{log});
            if ( -e $file )
            {
                my $numLines = 42;
                my @lines;
                eval {
                    @lines = ReadTailN($file, $numLines);
                };
                if ($@)
                {
                    print $fh $prefix."Failed to get last $numLines lines from $file: $@\n";
                }
                else
                {
                    print $fh $prefix."Last $numLines lines of $file:\n";
                    foreach my $line (@lines)
                    {
                        chomp($line);
                        print $fh $prefix.(' ' x 4)."[$line]\n";
                    }
                }
            }
            else
            {
                print $fh $prefix."Missing log file $file\n";
            }
        }
    }

    if ($haveLog and not $logToFile)
    {
        my $file = GetLogFileName('', $opts->{log}) || die;
        PrintResults($phases, $opts, $file);
    }

    return $errors;
}


sub BuildProcessDependencies($)
{
    my $phaseInfo = shift || die;

    if ( PhaseGetStatus($phaseInfo) )
    {
        return 0;
    }

    if ( $phaseInfo->{depends} )
    {
        foreach my $depInfo (@{$phaseInfo->{depends}})
        {
            my $status = PhaseGetStatus($depInfo);
            if ( not $status )
            {
                return 0;
            }

            if ($status eq $STATUS_ERROR)
            {
                PhaseSetStatus($phaseInfo, $STATUS_ERROR, $SUB_STATUS_DEP);
                return 0;
            }

            if ($status eq $STATUS_DONE)
            {
                next;
            }
            else
            {
                # this should never happen
                die "Unknown state";
            }
        }
    }

    return 1;
}


sub BuildActivePhasesSpawning($$)
{
    my $opts = shift || die;
    my $phases = shift || die;

    my $phaseList = $phases->{list} || die;
    my $phaseIndex = $phases->{index} || die;

    my @done;

    my $children = { count => 0 };

    my @phaseNames = @$phaseList;
    do
    {
        my @nextPhaseNames;

        foreach my $phaseName (@phaseNames)
        {
            my $phaseInfo = $phaseIndex->{$phaseName} || die;
            my $ready = BuildProcessDependencies($phaseInfo);

            #print "PHASE: $phaseName\n";

            if ( PhaseGetStatus($phaseInfo) )
            {
                push(@done, $phaseName);
            }
            elsif ( not $ready )
            {
                #print "NOT READY: $phaseName\n";
                push(@nextPhaseNames, $phaseName);
            }
            else
            {
                my $building = SpawnBuildPhase($opts, $phaseInfo, $children);
                if ( not $building )
                {
                    #print "NOT BUILDING: $phaseName\n";
                    push(@nextPhaseNames, $phaseName);
                }
            }
        }

        @phaseNames = @nextPhaseNames;

    } while (WaitForChild($children));

    if ($#phaseNames >= 0)
    {
        die "Did not finish the following phases: ".join(", ", @phaseNames)."\n";
    }
}


sub BuildActivePhases($$)
{
    my $opts = shift || die;
    my $phases = shift || die;

    my $phaseList = $phases->{list} || die;
    my $phaseIndex = $phases->{index} || die;

    print "Building phase list.\n";

    foreach my $phaseName (@$phaseList)
    {
        my $phaseInfo = $phaseIndex->{$phaseName} || die;
        if (not PhaseGetStatus($phaseInfo))
        {
            BuildPhase($opts, $phaseInfo);
        }
    }
}

########################################################################
#
# The code in main() cannot use any modules in the tree until after
# calling PrepareTree().
#

sub main()
{
    Getopt::Long::Configure('no_ignore_case', 'no_auto_abbrev') || die;

    sub GetDefaults()
    {
        my $opt = { debug => 1, optimize => 0 };

        eval
        {
            # This is strictly a convenience for testing the build script
            # from the tree where it resides such that we do not have
            # to pass in --repository and --revision in such cases.

            my $path = $FindBin::Bin;
            if ( -d File::Spec->catdir($path, '.svn') or
                 -d File::Spec->catdir($path, '_svn') )
            {
                foreach ( RunBackTick("$SVN info " . $FindBin::Bin ) )
                {
                    if ( /^URL:\s+(.*)\/build\s*$/ ) {
                        $opt->{repository} = $1;
                    }
                    elsif ( /^Revision: (\d*)$/ ) {
                        $opt->{revision} = $1;
                    }
                }
            }
        };
        if ( $@ )
        {
            delete $opt->{root};
            delete $opt->{repository};
            delete $opt->{revision};
        }

        return $opt;
    }

    my $opt = GetDefaults();
    my $ok = GetOptions($opt,
                        'help|h|?',
                        'test',
                        'revert',
                        'log=s',
                        'nospawn',
                        'noprepare',
                        'nobuildenv',

                        'config=s',
                        'phase=s',
                        'define|D=s@',

                        'name=s',
                        'number=s',
                        'repository=s',
                        'revision=s',

                        'user=s',
                        'scp_command=s',
                        'ssh_command=s',

                        'machine=s',
                        'scpdist=s',
                        'root=s',
                        'dist=s',
                        'image=s',
			'meta=s',
			'cache=s',
			'run=s',
                        'path=s',
                        'command=s',
                        'checkdist=s',
                        'clean_dist',
                        'sparse!',

                        'debug!',
                        'optimize!',
                       );

    my $errors = '';
    foreach my $optionName (@REQUIRED_OPTIONS_KEYS)
    {
        if ( ( not defined ($opt->{$optionName}) ) or
             ( '' eq $opt->{$optionName} ) )
        {
            $errors .= "Missing value for required option --".$optionName.".\n";
        }
    }

    if ( not ($opt->{config} xor $opt->{phase}) )
    {
        $errors .= "Exactly one of --config or --phase must be specified.\n";
    }
    else
    {
        if ( $opt->{phase} )
        {
            if ($opt->{define})
            {
                $errors .= "Cannot use --define with --phase.\n";
            }
            foreach my $optionName qw(root dist user)
            {
                if ( ( not defined ($opt->{$optionName}) ) or
                     ( '' eq $opt->{$optionName} ) ) {
                    $errors .= "Option --".$optionName." is required with --phase.\n";
                }
            }
            my @invalid = qw(log nospawn define);
            if ($opt->{machine})
            {
                if (not ($opt->{machine} =~ /^[A-Z0-9_]/i))
                {
                    $errors .= "Invalid --machine value: $opt->{machine}\n"
                }
            }
            else
            {
                push(@invalid, qw(ssh_command scp_command scpdist));
            }
            foreach my $optionName (@invalid)
            {
                if ( defined ($opt->{$optionName}) )
                {
                    $errors .= "Option --".$optionName." is not allowed with --phase.\n";
                }
            }
        }
        else
        {
            die if not $opt->{config};
            if ( not ($opt->{log} xor $opt->{nospawn}) )
            {
                $errors .= "Exactly one of --log or --nospawn must be specified with --config.\n";
            }
            foreach my $optionName qw(machine scpdist root dist path command checkdist clean_dist sparse nosparse)
            {
                if ( defined ($opt->{$optionName}) )
                {
                    $errors .= "Option --".$optionName." is not allowed with --config.\n";
                }
            }
        }
    }

    my $defs = {};
    foreach my $define (@{$opt->{define}})
    {
        if ($define =~ /^($ID_PATTERN)\s*=\s*(.*)$/)
        {
            my $key = $1;
            my $value = $2;
            if ( exists $defs->{$key} )
            {
                $errors .= "Multiple --define options for $key\n";
            }
            else
            {
                $defs->{$key} = $value;
            }
        }
        else
        {
            $errors .= "Invalid --define option: \"$define\"\n";
        }
    }

    if ( $#ARGV != -1 )
    {
        $errors .= "Unexpected argument: $ARGV[0]\n";
    }

    if ( $opt->{help} )
    {
        print usage();
        exit(1);
    }
    elsif ( not $ok or $errors )
    {
        die $errors . "For usage information, use the --help option.\n";
    }

    $TEST_MODE = 1 if $opt->{test};

    if ($opt->{log})
    {
        GetLogFileName(undef, $opt->{log}, 1);
    }

    if (not $opt->{phase})
    {
        my $argOpts = {};
        map {
            if ( exists $opt->{$_} )
            {
                $argOpts->{$_} = $opt->{$_};
            }
        } @VALID_OPTIONS_KEYS;

        print "Reading configuration from $opt->{config}.\n";
        my $configInfo = ReadConfig( $opt->{config}, $argOpts, $defs );
        print "Finished reading configuration.\n\n";

        my $phases = $configInfo->{phases} || die;

        $opt->{user} = $configInfo->{opts}->{user} || die;
        $opt->{scp_command} = $configInfo->{opts}->{scp_command} || die;
        $opt->{ssh_command} = $configInfo->{opts}->{ssh_command} || die;

        CheckSkip($phases);
        CheckConnections($opt->{ssh_command}, $phases);

        {
            print "\nGoing to build the following phases:\n";
            my $phaseList = $phases->{list} || die;
            my $phaseIndex = $phases->{index} || die;
            foreach my $phaseName (@$phaseList)
            {
                my $phaseInfo = $phaseIndex->{$phaseName} || die;
                my $machine = $phaseInfo->{machine} || die;
                print "  $phaseName @ $machine\n" if not PhaseGetStatus($phaseInfo);
            }
        }

        eval {
            if ($opt->{nospawn})
            {
                BuildActivePhases($opt, $phases);
            }
            else
            {
                BuildActivePhasesSpawning($opt, $phases);
            }
        };
        if ($@)
        {
            print STDERR "$@";
            PrintResults($phases, $opt);
            return 1;
        }
        my $errors = PrintResults($phases, $opt);
        return $errors;
    }
    else
    {
        print "Building phase: $opt->{phase}\n";
        if ($opt->{machine})
        {
            my $phaseInfo = { name => $opt->{phase} };
            # Ideally, should use subset of VALID_PHASE_KEYS,
            # but we rely on opt checking to weed out stuff we
            # do not want.
            map {
                if (exists $opt->{$_})
                {
                    $phaseInfo->{$_} = $opt->{$_};
                }
            } @VALID_PHASE_KEYS;
            if ($phaseInfo->{scpdist})
            {
                $phaseInfo->{scpdist} = EvaluateScpDist($phaseInfo->{scpdist}, 'N/A');
            }
            BuildRemote($opt, $phaseInfo);
        }
        else
        {
            BuildLocal($opt);
        }
    }

    return 0;
}
