/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright Likewise Software
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.  You should have received a copy of the GNU General
 * Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        fileFsAttribInfo
 *
 * Abstract:
 *
 *        Likewise Posix File System Driver (PVFS)
 *
 *        FileFsAttributeInformation Handler
 *
 * Authors: Gerald Carter <gcarter@likewise.com>
 */

#include "pvfs.h"

/* Forward declarations */

static
NTSTATUS
PvfsQueryFileFsAttribInfo(
    PPVFS_IRP_CONTEXT pIrpContext
    );


/* File Globals */



/* Code */


NTSTATUS
PvfsFileFsAttribInfo(
    PVFS_INFO_TYPE Type,
    PPVFS_IRP_CONTEXT pIrpContext
    )
{
    NTSTATUS ntError = STATUS_UNSUCCESSFUL;

    switch(Type)
    {
    case PVFS_SET:
        ntError = STATUS_NOT_SUPPORTED;
        break;

    case PVFS_QUERY:
        ntError = PvfsQueryFileFsAttribInfo(pIrpContext);
        break;

    default:
        ntError = STATUS_INVALID_PARAMETER;
        break;
    }
    BAIL_ON_NT_STATUS(ntError);

cleanup:
    return ntError;

error:
    goto cleanup;
}


static
NTSTATUS
PvfsQueryFileFsAttribInfo(
    PPVFS_IRP_CONTEXT pIrpContext
    )
{
    NTSTATUS ntError = STATUS_UNSUCCESSFUL;
    PIRP pIrp = pIrpContext->pIrp;
    PPVFS_CCB pCcb = NULL;
    PFILE_FS_ATTRIBUTE_INFORMATION pFileInfo = NULL;
    const IRP_ARGS_QUERY_SET_VOLUME_INFORMATION* pArgs =
        &pIrpContext->pIrp->Args.QuerySetVolumeInformation;
    PWSTR pwszFsName = NULL;
    size_t FsNameLenBytes = RtlCStringNumChars(PVFS_FS_NAME) * sizeof(WCHAR);
    size_t FixedSize =
         LW_FIELD_OFFSET(FILE_FS_ATTRIBUTE_INFORMATION, FileSystemName);
    PVFS_STATFS StatFs = {0};
    ULONG FsAttributeSet = PVFS_FILE_SYSTEM_ATTRIBUTES;

    if (gPvfsDriverState.Config.EnableEaSupport)
    {
        SetFlag(FsAttributeSet, FILE_SUPPORTS_EXTENDED_ATTRIBUTES);
    }

    /* Sanity checks */

    ntError = PvfsAcquireCCB(pIrp->FileHandle, &pCcb);
    BAIL_ON_NT_STATUS(ntError);

    /* not sure exact access rights to check for here */

    ntError = PvfsAccessCheckFileHandle(pCcb, FILE_READ_ATTRIBUTES);
    BAIL_ON_NT_STATUS(ntError);

    BAIL_ON_INVALID_PTR(pArgs->FsInformation, ntError);

    if (pArgs->Length < FixedSize)
    {
        ntError = STATUS_BUFFER_TOO_SMALL;
        BAIL_ON_NT_STATUS(ntError);
    }

    pFileInfo = (PFILE_FS_ATTRIBUTE_INFORMATION)pArgs->FsInformation;

    /* Real work starts here */

    ntError = PvfsSysFstatFs(pCcb, &StatFs);
    BAIL_ON_NT_STATUS(ntError);

    ntError = RtlWC16StringAllocateFromCString(&pwszFsName, PVFS_FS_NAME);
    BAIL_ON_NT_STATUS(ntError);

    pFileInfo->FileSystemAttributes = FsAttributeSet;
    pFileInfo->MaximumComponentNameLength = StatFs.MaximumNameLength;
    pFileInfo->FileSystemNameLength = FsNameLenBytes;

    if (FsNameLenBytes > (pArgs->Length - FixedSize))
    {
        FsNameLenBytes = pArgs->Length - FixedSize;
        ntError = STATUS_BUFFER_OVERFLOW;
    }

    memcpy(pFileInfo->FileSystemName, pwszFsName, FsNameLenBytes);

    pIrp->IoStatusBlock.BytesTransferred = FixedSize + FsNameLenBytes;

cleanup:
    if (pCcb)
    {
        PvfsReleaseCCB(pCcb);
    }

    PVFS_FREE(&pwszFsName);

    return ntError;

error:
    goto cleanup;
}
