/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright Likewise Software
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.  You should have received a copy of the GNU General
 * Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        context.c
 *
 * Abstract:
 *
 *        Likewise IO (LWIO) - SRV
 *
 *        Elements
 *
 *        Execution Context
 *
 * Authors: Sriram Nambakam (snambakam@likewise.com)
 *
 */

#include "includes.h"

static
VOID
SrvFreeExecContext(
   IN PSRV_EXEC_CONTEXT pContext
   );

NTSTATUS
SrvBuildExecContext(
   IN  PLWIO_SRV_CONNECTION pConnection,
   IN  PSMB_PACKET          pSmbRequest,
   IN  BOOLEAN              bInternal,
   OUT PSRV_EXEC_CONTEXT*   ppContext
   )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PSRV_EXEC_CONTEXT pContext = NULL;

    ntStatus = SrvAllocateMemory(sizeof(*pContext), OUT_PPVOID(&pContext));
    BAIL_ON_NT_STATUS(ntStatus);

    pContext->refCount = 1;
    pContext->bInternal = bInternal;
    pContext->bInline = TRUE;

    ntStatus = LwErrnoToNtStatus(pthread_mutex_init(&pContext->execMutex, NULL));
    BAIL_ON_NT_STATUS(ntStatus);
    pContext->pExecMutex = &pContext->execMutex;

    ntStatus = LwRtlCreateWorkItem(
                    gSrvElements.ThreadPool,
                    &pContext->pWorkItem,
                    gSrvElements.WorkItemFunction,
                    pContext);
    BAIL_ON_NT_STATUS(ntStatus);

    ntStatus = SrvLogContextCreate(&pContext->pLogContext);
    BAIL_ON_NT_STATUS(ntStatus);

    ntStatus = SrvLogContextUpdateFilter(
                    pContext->pLogContext,
                    pConnection->pClientAddress,
                    pConnection->clientAddrLen);
    BAIL_ON_NT_STATUS(ntStatus);

    pContext->pConnection = SrvConnectionAcquire(pConnection);

    pContext->pSmbRequest = pSmbRequest;
    if (pSmbRequest)
    {
        InterlockedIncrement(&pSmbRequest->refCount);
    }

cleanup:

    *ppContext = pContext;

    return ntStatus;

error:

    if (pContext)
    {
        SrvReleaseExecContext(pContext);
        pContext = NULL;
    }

    goto cleanup;
}

BOOLEAN
SrvIsValidExecContext(
   IN PSRV_EXEC_CONTEXT pExecContext
   )
{
    return (pExecContext &&
            pExecContext->pConnection &&
            pExecContext->pSmbRequest);
}

VOID
SrvReleaseExecContextHandle(
   IN HANDLE hExecContext
   )
{
    SrvReleaseExecContext((PSRV_EXEC_CONTEXT)hExecContext);
}

PSRV_EXEC_CONTEXT
SrvAcquireExecContext_impl(
   PSRV_EXEC_CONTEXT pContext,
   PCSTR func,
   LONG line
   )
{
    LW_LONG ref = InterlockedIncrement(&pContext->refCount);

    LWIO_LOG_DEBUG("Acquiring exec context [%p, ref:%d] %s()%d",
            pContext, 
            ref, 
            func, 
            line);

    return pContext;
}

VOID
SrvReleaseExecContext_impl(
   IN PSRV_EXEC_CONTEXT pContext,
   PCSTR func,
   LONG line
   )
{
    LW_LONG ref = InterlockedDecrement(&pContext->refCount);

    LWIO_LOG_DEBUG("Releasing exec context [%p, ref:%d] %s()%d",
            pContext,
            ref,
            func,
            line);

    if (ref == 0)
    {
        SrvFreeExecContext(pContext);
    }
}

static
VOID
SrvFreeExecContext(
   IN PSRV_EXEC_CONTEXT pContext
   )
{
    SrvMpxTrackerRemoveExecContext(pContext);

    if (pContext->pAcb && pContext->pAcb->AsyncCancelContext)
    {
        IoDereferenceAsyncCancelContext(&pContext->pAcb->AsyncCancelContext);
    }

    if (pContext->pProtocolContext)
    {
        pContext->pfnFreeContext(pContext, pContext->pProtocolContext);
        pContext->pProtocolContext = NULL;
    }

    if (pContext->pSmbRequest)
    {
        SMBPacketRelease(
            pContext->pConnection->hPacketAllocator,
            pContext->pSmbRequest);
    }

    if (pContext->pSmbResponse)
    {
        SMBPacketRelease(
            pContext->pConnection->hPacketAllocator,
            pContext->pSmbResponse);
    }

    if (pContext->pInterimResponse)
    {
        SMBPacketRelease(
            pContext->pConnection->hPacketAllocator,
            pContext->pInterimResponse);
    }

    if (pContext->pConnection)
    {
        SrvConnectionRelease(pContext->pConnection);
    }

    if (pContext->pStatInfo)
    {
        SrvStatisticsRelease(pContext->pStatInfo);
    }

    if (pContext->pLogContext)
    {
        SrvLogContextFree(pContext->pLogContext);
    }

    LwRtlFreeWorkItem(&pContext->pWorkItem);

    if (pContext->pExecMutex)
    {
        pthread_mutex_destroy(&pContext->execMutex);
    }

    SrvFreeMemory(pContext);
}

